#include <stdio.h>
#include <ctype.h>
#include <string.h>

extern "C" {
  #include "LeddarC.h"
  #include "LeddarProperties.h"
}
#include "serial.h"

#define ARRAY_LEN( a )  (sizeof(a)/sizeof(a[0]))

#ifndef M_PI
#    define M_PI 3.14159265358979323846
#endif

// Global variable to avoid passing to each function.
static LeddarHandle gHandle=NULL;
FILE *fo=NULL;
volatile int vertices = 0;
volatile int angle = 0;
volatile int finish = 0;
volatile int busy = 0;
Serial serial;


static void CheckError( int aCode )
{
    if ( aCode != LD_SUCCESS )
    {        
        printf( "LeddarC error (%d): ", aCode );

        switch ( aCode )
        {
            case LD_ACCESS_DENIED : printf("LD_ACCESS_DENIED\n" ); break;
            case LD_TIMEOUT : printf("LD_TIMEOUT\n" ); break;
            case LD_START_OF_FILE : printf("LD_START_OF_FILE\n" ); break;
            case LD_END_OF_FILE : printf("LD_END_OF_FILE\n" ); break;
            case LD_NO_RECORD : printf("LD_NO_RECORD\n" ); break;
            case LD_ALREADY_STARTED : printf("LD_ALREADY_STARTED\n" ); break;
            case LD_NO_DATA_TRANSFER : printf("LD_NO_DATA_TRANSFER\n" ); break;
            case LD_NOT_CONNECTED : printf("LD_NOT_CONNECTED\n" ); break;
            case LD_INVALID_ARGUMENT : printf("LD_INVALID_ARGUMENT\n" ); break;
            case LD_ERROR : printf("LD_ERROR\n" ); break;
            case LD_NOT_ENOUGH_SPACE : printf("LD_NOT_ENOUGH_SPACE\n" ); break;
        }
    }
}

static char WaitKey( void )
{
    while( !LeddarKeyPressed() )
    {
        LeddarSleep( 0.5 );
    }

    return toupper( LeddarGetKey() );
}

static void DataCallback( void *aHandle )
{    
	fflush(fo);
	if (busy) return;
	busy = 1;
	LdDetection lDetections[16];
	Sleep(20);

    unsigned int i, j, lCount = LeddarGetDetectionCount( aHandle );

    if ( lCount > ARRAY_LEN( lDetections ) )
    {
        lCount = ARRAY_LEN( lDetections );
    }

    LeddarGetDetections( aHandle, lDetections, ARRAY_LEN( lDetections ) );

	if (angle >= 180) finish=1;	

    // When replaying a record, display the current index
    if ( LeddarGetRecordSize( gHandle ) != 0 )
    {
        printf( "%3d, %6d: ", angle, LeddarGetCurrentRecordIndex( gHandle ) );
    }
		
	if (finish==0) {
	  printf("%3d: ", angle);	
	  for (i = 0; i < 16; ++i)
	  {
		  /*lDetections[i].mSegment*/
		  float dist = 30;
		  if (i < lCount){
			  dist = lDetections[i].mDistance;
			  if (dist > 30) dist = 30;
		  }
		  else {
			  dist = lDetections[i-1].mDistance;
		  }
		  if (abs(dist < 0.01)) dist = 0.01;		  
		  printf("%5.2f ", dist);
		  float lidarangle = (((float)i)*2.8125 - 22.5) / 180.0 * M_PI;
		  float groundangle = ((float)-angle) / 180.0 * M_PI;
		  float grounddist = cos(lidarangle) * dist;
		  float x = cos(groundangle)*grounddist;
		  float y = sin(groundangle)*grounddist;
		  float z = sin(lidarangle)*dist;		
	      fprintf(fo, "%.2f %.2f %.2f 255 0 0\n", x, y, z);		 
		  vertices++;
	  }
	  puts("");
	  angle++;

	  char buf[100];
	  sprintf(buf, "!01,%d\r\n", angle);
	  serial.print(buf);
    }    
	
	busy = 0;
}



static void ReplayData( void )
{
    puts( "\nP to go forward, O to go backward, H to return to beginning, Q to quit" );

    CheckError( LeddarStartDataTransfer( gHandle, LDDL_DETECTIONS ) );
    LeddarSetCallback( gHandle, DataCallback, gHandle );

    for(;;)
    {
        char lChoice = WaitKey();

        switch( lChoice )
        {
            case 'H':
                LeddarMoveRecordTo( gHandle, 0 );
                break;
            case 'O':
                CheckError( LeddarStepBackward( gHandle ) );
                break;
            case 'P':
                CheckError( LeddarStepForward( gHandle ) );
                break;
            case 'Q':
            case  27: // Escape
                LeddarStopDataTransfer( gHandle );
                LeddarRemoveCallback( gHandle, DataCallback, gHandle );
                return;
        }
    }
}


static void
ReadConfiguration( void )
{
    double lValue;
    char   lValueStr[64];

    puts( "\nCurrent Configuration:\n" );
    CheckError( LeddarGetTextProperty( gHandle, PID_NAME, 0, lValueStr ) );
    printf( "  Device Name     : %s\n", lValueStr );
    CheckError( LeddarGetProperty( gHandle, PID_BASE_POINT_COUNT, 0, &lValue ) );
    printf( "  Base Point Count: %.0f\n", lValue );
    CheckError( LeddarGetProperty( gHandle, PID_LED_INTENSITY, 0, &lValue ) );
    printf( "  Led Intensity   : %.0f\n", lValue );
    CheckError( LeddarGetProperty( gHandle, PID_THRESHOLD_OFFSET, 0, &lValue ) );
    printf( "  Threshold offset: %.2f\n", lValue );
    
    puts( "\nPress a key to continue." );
    WaitKey();
}



static void
ListSensors( char* aConnectyionType, char* aAddresses, unsigned int aSize )
{
    char         lConnectionType[256];
    unsigned int lIndex = 0;

    if ( aConnectyionType == NULL )
    {
        printf("\nEnter the connection type (USB or SERIAL): ");
        scanf("%255s", lConnectionType );
    }
    else
    {
        strcpy( lConnectionType, aConnectyionType );
    }

    strcpy( aAddresses, aConnectyionType );

    CheckError( LeddarListSensors( aAddresses, &aSize ) );

    printf( "\nFound %d sensors of type %s\n", aSize, lConnectionType );

    int lConnectionFoundIndex = 0;
    while( strlen( aAddresses+lIndex ) > 0 )
    {
        if ( lConnectionFoundIndex % 2 == 0 )
        {
            if ( strcmp( aConnectyionType, "USB" ) != 0 )
            {
                printf( "%d : %s", lConnectionFoundIndex/2, aAddresses + lIndex );
            }
            else
            {
                printf( "%d : ", lConnectionFoundIndex/2 );
            }
        }
        else
        {
            if ( strcmp( aConnectyionType, "USB" ) != 0 )
            {
                printf( "\t[ %s ]\n", aAddresses + lIndex );
            }
            else
            {
                printf( "[ %s ]\n", aAddresses + lIndex );
            }
        }

        lConnectionFoundIndex++;
        lIndex += strlen( aAddresses+lIndex ) + 1;
    }
}


static void
ConfigurationMenu( void )
{
    while( LeddarGetConnected( gHandle ) == LD_SUCCESS )
    {
        char         lChoice;
        unsigned int lId = 0;
        unsigned int lType = 1;

        puts( "\nConfiguration Change Menu" );
        puts( "  1. Change Oversampling Exponent" );
        puts( "  2. Change Accumulation Exponent" );
        puts( "  3. Change Base Point Count" );
        puts( "  4. Change Led Intensity" );
        puts( "  5. Change Threshold Offset" );
        puts( "  6. Change Name" );
        puts( "  7. Write" );
        puts( "  8. Restore" );
        puts( "  9. Quit" );

        lChoice = WaitKey();

        switch( lChoice )
        {
            case '1':
                lId = PID_OVERSAMPLING_EXPONENT;
                break;
            case '2':
                lId = PID_ACCUMULATION_EXPONENT;
                break;
            case '3':
                lId = PID_BASE_POINT_COUNT;
                break;
            case '4':
                lId = PID_LED_INTENSITY;
                break;
            case '5':
                lId = PID_THRESHOLD_OFFSET;
                break;
            case '6':
                lId = PID_NAME;
                lType = 2;
                break;
            case '7':
                CheckError( LeddarWriteConfiguration( gHandle ) );
                break;
            case '8':
                CheckError( LeddarRestoreConfiguration( gHandle ) );
                break;
            case '9':
            case  27: // Escape
                if ( LeddarGetConfigurationModified( gHandle ) != LD_SUCCESS )
                {
                    return;
                }

                puts( "\n** Configuration modified, please Write or Restore before quitting **" );
                break;
        }

        if ( lId != 0 )
        {
            printf( "\nEnter new value: " );

            switch( lType )
            {
                case 1:
                {
                    double lValue;

                    scanf( "%lf", &lValue );
                    CheckError( LeddarSetProperty( gHandle, lId, 0, lValue ) );
                }
                    break;
                case 2:
                {
                    char lValue[64];

                    scanf( "%63s", lValue );
                    CheckError( LeddarSetTextProperty( gHandle, lId, 0, lValue ) );
                }
                    break;
            }
        }
    }
}


static char* 
FindAddressByIndex( unsigned int aIndex, char* aAddresses )
{
    unsigned int lCurrentIndex = 0;
    unsigned int lConnectionFoundIndex = 0;

    while( strlen( aAddresses+lCurrentIndex ) > 0 )
    {
        if ( ( lConnectionFoundIndex / 2 ) == aIndex )
        {
            return aAddresses + lCurrentIndex;
        }

        lConnectionFoundIndex++;
        lCurrentIndex += strlen( aAddresses+lCurrentIndex ) + 1;
    }

    return NULL;
}


static void
ConnectMenu( int aTrySingleUsb )
{
    char lAddresses[256];
    char* lAddress = NULL;
    char lConnectionType[10];

    if ( aTrySingleUsb )
    {
        strcpy( lConnectionType, "USB" );
        ListSensors( lConnectionType, lAddresses, 255 );
        lAddress = FindAddressByIndex( 0, lAddresses );
        if ( lAddress == NULL )
        {
            return;
        }
    }
    else
    {
        // Ask for connection type and try to connect before displaying menu.
        printf( "\nEnter connection type (USB or SERIAL): " );
        scanf( "%24s", lConnectionType );    

        ListSensors( lConnectionType, lAddresses, 255 );

        unsigned int lSensorNumber = 0;
        printf("\nConnect to sensor number: ");
        scanf("%u", &lSensorNumber );
        
        lAddress = FindAddressByIndex( lSensorNumber, lAddresses );
        if ( lAddress == NULL )
        {
            printf("\nError, sensor index not valid.");
            return;
        }
    }

    if ( LeddarConnect( gHandle, lConnectionType, lAddress ) == LD_SUCCESS )
    {
        while( LeddarGetConnected( gHandle ) == LD_SUCCESS )
        {
            char lChoice;

            puts( "\n\nConnected Menu" );
            puts( "  1. Read Data" );
            puts( "  2. Read Configuration" );
            puts( "  3. Change Configuration" );
            if ( LeddarGetRecording( gHandle ) == LD_SUCCESS )
            {
                puts( "  4. Stop Recording" );
            }
            else
            {
                puts( "  4. Start Recording" );
            }
            puts( "  5. Disconnect" );

            lChoice = WaitKey();

            switch( lChoice )
            {
                case '1':
                    //ReadLiveData();
                    break;
                case '2':
                    ReadConfiguration();
                    break;
                case '3':
                    ConfigurationMenu();
                    break;
                case '4':
                    if ( LeddarGetRecording( gHandle ) == LD_SUCCESS )
                    {
                        LeddarStopRecording( gHandle );
                    }
                    else
                    {
                        LeddarChar lFilename[255];
                        CheckError( LeddarStartRecording( gHandle, lFilename ) );
                    }
                    break;
                case '5':
                case  27:
                    LeddarDisconnect( gHandle );
                    return;
            }
        }
    }
    else
    {
        puts( "\nConnection failed!" );
    }    
}


static void
ReplayMenu( void )
{
    LeddarChar lName[256];

    // Ask for file name and try to load record before display menu.
    printf( "\nEnter file name: " );
    scanf( "%255s", lName );

    if ( LeddarLoadRecord( gHandle, lName ) == LD_SUCCESS )
    {
        puts( "\nPlease wait while the record is loading..." );

        // For a big file, especially if it is on a network drive, it may
        // take a while before the replay is 100% ready. Note that you
        // can still use the replay but it will not report the complete
        // size until it is finished loading.
        while( LeddarGetRecordLoading( gHandle ) )
        {
            LeddarSleep( 0.5 );
        }

        printf( "Finished loading record of %d frames.\n", 
                LeddarGetRecordSize( gHandle ) );

        for(;;)
        {
            char lChoice;

            puts( "\nReplay Menu" );
            puts( "  1. Read Data" );
            puts( "  2. Read Configuration" );
            puts( "  3. Close" );

            lChoice = WaitKey();

            switch( lChoice )
            {
                case '1':
                    ReplayData();
                    break;
                case '2':
                    ReadConfiguration();
                    break;
                case '3':
                case  27:
                    LeddarDisconnect( gHandle );
                    return;
            }
        }
    }
    else
    {
        puts( "\nFailed to load file!" );
    }
}


static void
ConfigureRecordingMenu( void )
{
    for(;;)
    {
int    lChoice;
LeddarChar lPath[256];

puts("\nConfigure Recording Menu");
LeddarGetRecordingDirectory(lPath, ARRAY_LEN(lPath));
printf("  1. Change directory (%s)\n", lPath);
printf("  2. Change max file size (%dMB)\n", LeddarGetMaxRecordFileSize());
puts("  3. Quit");

lChoice = toupper(LeddarGetKey());

switch (lChoice)
{
case '1':
	printf("\nEnter recording directory: ");
	//LeddarScanf( LTS( "%255s" ), lPath );
	scanf("%255s", lPath);
	LeddarConfigureRecording(lPath, 0, 0);
	break;
case '2':
	printf("\nEnter max file size in MB: ");
	scanf("%d", &lChoice);
	LeddarConfigureRecording(NULL, lChoice, 0);
	break;
case '3':
case  27:
	return;
}
	}
}


static void
MainMenu(void)
{

	for (;;)
	{
		int lChoice;

		puts("\n\nMain Menu");
		puts("  1. Connect");
		puts("  2. Connect to single USB sensor");
		puts("  3. List Sensors");
		puts("  4. Replay Record");
		puts("  5. Configure Recording");
		puts("  6. Quit");

		lChoice = toupper(LeddarGetKey());

		switch (lChoice)
		{
		case '1':
			ConnectMenu(0);
			break;
		case '2':
			ConnectMenu(1);
			break;
		case '3':
		{
			char lAddresses[256];
			ListSensors("USB", lAddresses, 255);
			ListSensors("SERIAL", lAddresses, 255);
			break;
		}
		case '4':
			ReplayMenu();
			break;
		case '5':
			ConfigureRecordingMenu();
			break;
		case '6':
		case 'Q':
		case  27:
			puts("\n*** Goodbye! ***");
			return;
		}
	}
}


static void ReadLiveData(void)
{
	//puts("\nPress a key to start reading data and press a key again to stop.");
	//WaitKey();
	char buf[100];
	
	Sleep(2000);
	puts("start");
	serial.print("!01,0\r\n");
	Sleep(2000);

	CheckError(LeddarStartDataTransfer(gHandle, LDDL_DETECTIONS));
	LeddarSetCallback(gHandle, DataCallback, gHandle);

	//WaitKey();
	//puts("start\n");	
	while (!finish) {						
		Sleep(10);
	}	
	while (busy) {
		Sleep(10);
	}
	puts("done\n");	

	LeddarStopDataTransfer(gHandle);
	LeddarRemoveCallback(gHandle, DataCallback, gHandle);
}



int main(int argc, char *argv[])
{
	puts("Lidar");
	serial.begin("\\\\.\\com12", 115200);	

	fo = fopen("lidar.ply", "w");
	fprintf(fo, "ply\n");
    fprintf(fo, "format ascii 1.0\n");
	fprintf(fo, "element vertex 2880\n");
	fprintf(fo, "property float x\n");
	fprintf(fo, "property float y\n");
	fprintf(fo, "property float z\n");
	fprintf(fo, "property uchar red\n");
	fprintf(fo, "property uchar green\n");
	fprintf(fo, "property uchar blue\n");
	fprintf(fo, "element face 2685\n");
	fprintf(fo, "property list uchar int vertex_indices\n");
	fprintf(fo, "end_header\n");	
	//fprintf(fo, "0.0 0.0 0.0 255 0 0\n");
	
	gHandle = LeddarCreateWithConsole(argc, argv);

	//MainMenu();
	char* lAddress = NULL;
	char         lConnectionType[256];
	strcpy(lConnectionType, "USB");
	char lAddresses[256];
	ListSensors(lConnectionType, lAddresses, 255);
	lAddress = FindAddressByIndex(0, lAddresses);
	if (lAddress == NULL)
	{
		printf("\nError, sensor index not valid.");
		return 1;
	}
	if (!LeddarConnect(gHandle, lConnectionType, lAddress) == LD_SUCCESS)
	{
		printf("\nError connecting");
		return 1;
	}
		
	ReadLiveData();	

	for (int i = 0; i < 179; i++){				
		for (int j = 0; j < 15; j++){
		  int a = i * 16 + j+1;
		  int b = i * 16 + j;
		  int c = b + 16;
		  int d = a + 16;
		  fprintf(fo, "4 %d %d %d %d\n", a,b,c,d);		  		  
		}		
	}
			
	fclose(fo);
    LeddarDestroy( gHandle );
	printf("vertices %d\n", vertices);

    return 0;
}


